package com.biganto.visual.roompark.domain.use_case

import android.content.res.Resources
import com.biganto.visual.roompark.data.repository.api.biganto.IBigantoApi
import com.biganto.visual.roompark.data.repository.db.IDb
import com.biganto.visual.roompark.domain.custom_exception.CustomApiException
import com.biganto.visual.roompark.domain.model.TourModel
import com.biganto.visual.roompark.domain.model.fromEntity
import com.biganto.visual.roomparkvr.data.repository.db.requery.model.DownloadState
import com.biganto.visual.roomparkvr.data.repository.db.requery.model.TourPreviewEntity
import com.biganto.visual.roomparkvr.data.repository.db.requery.model.fromRaw
import io.reactivex.Observable
import io.reactivex.schedulers.Schedulers
import timber.log.Timber
import javax.inject.Inject

/**
 * Created by Vladislav Bogdashkin on 19.06.2018.
 */

const val EMPTY_PARENT = -778

class TourPreviewsUseCase @Inject constructor(
    private val api:IBigantoApi
    ,private val db: IDb
) {

    fun getTourById(tourId:String):Observable<DownloadTourByIdResult> {
        val tour = retrieveTour(tourId)
        Timber.d("retrieveTour $tourId / $tour")
        if (tour != null) {
            if (tour.downloadState == DownloadState.Downloaded)
                return Observable.just<DownloadTourByIdResult>(DownloadTourByIdResult.ReadyToPlay(tour))

            return setTourState(tour.tour_id, DownloadState.DownloadQueue)
                .map<DownloadTourByIdResult> { DownloadTourByIdResult.DownloadStarted(tour) }
        }

        return syncTour(tourId, parentId = EMPTY_PARENT)
            .map { it }
            .map<DownloadTourByIdResult> { DownloadTourByIdResult.DownloadStarted(it) }
    }

    fun retrieveTour(tourId: String) =
            db.getTourPreview(tourId)
                .asSequence()
                .map ( ::fromEntity)
                .firstOrNull()

    private fun setTourState(tourId:String, state: DownloadState): Observable<Boolean> =
                    db.getTourPreview(tourId)
                        .observable()
                        .doOnNext{ it.isDownloaded=state }
                        .flatMap { db.upsertTourPreview(it)?.map{true} }

    private fun syncTour(tourId:String, parentId:Int):Observable<TourModel> =
        syncTour(tourId,parentId,calcTargetResolution())

    private fun syncTour(tourId:String, estateId: Int, targetResolution:Int)
            : Observable<TourModel> {
        val request = ToursRemoteByIdRequestModel(
            arrayListOf(tourId)
            , estateId
            , null
            , targetResolution
        )
        return forceTourUpdate(request)
            .map {
                if (it.isNullOrEmpty())
                    throw CustomApiException.TourByIdNotFoundException()
                it.first() }
            .doOnNext { it.isDownloaded = DownloadState.DownloadQueue }
            .flatMap {db.upsertTourPreview(it)}
            .map(::fromEntity)
            .doOnError(Timber::e)
    }


    private fun calcTargetResolution():Int{
        val w = Resources.getSystem().displayMetrics.widthPixels
        val h = Resources.getSystem().displayMetrics.heightPixels
        return w.coerceAtLeast(h)
    }

    private fun forceTourUpdate(requestModel: ToursRemoteByIdRequestModel): Observable<List<TourPreviewEntity>> =
        api.getToursPreviewById(requestModel.ids)
            .subscribeOn(Schedulers.io())
            .map { fromRaw(it
                , requestModel.estateId
                , api.provideHttpUrl().toString()
                , requestModel.targetResolution)
            }
}

data class TourRemoteRequestModel(val estateId: String, val token: String? = null, val targetResolution: Int)


data class ToursRemoteByIdRequestModel(val ids: List<String>, val estateId: Int,
                                       val token: String? = null,  val targetResolution: Int)


sealed class DownloadTourByIdResult{
    class ReadyToPlay(val tour:TourModel): DownloadTourByIdResult()
    class DownloadStarted(val tour:TourModel): DownloadTourByIdResult()
    class TourError(val id:String): DownloadTourByIdResult()

}
